/*-
 * This file is a part of the MDCacheD server
 * (c) 2007.-2010. Ivan Voras <ivoras@gmail.com>
 * Released under the 2-clause BSDL
 * $Id: Mutex.h 347 2011-09-20 20:58:41Z ivoras $
 */

#ifndef _MUTEX_H_
#define _MUTEX_H_

#include <pthread.h>
#include <exception>


class MutexException : public std::exception {
private:
	char *msg;
public:
	MutexException(char *msg) : exception() {
		this->msg = msg;
	}
};


class Mutex {
public:
	Mutex() throw (MutexException) {
		if (pthread_mutex_init(&mtx, NULL) != 0)
			throw new MutexException((char*)"pthread_mutex_init failed at " __FILE__);
	}
	~Mutex() {
		pthread_mutex_destroy(&mtx);
	}
	void Lock() {
		pthread_mutex_lock(&mtx);
	}
	bool TryLock() {
		return pthread_mutex_trylock(&mtx) == 0;
	}
	void Unlock() {
		pthread_mutex_unlock(&mtx);
	}
private:
	pthread_mutex_t mtx;
};

#endif
