/*-
 * This file is a part of the MDCacheD server
 * (c) 2012. Ivan Voras <ivoras@gmail.com>
 * Released under the 2-clause BSDL
 * $Id: ServerModelMetadata.h 452 2012-03-20 16:03:25Z ivoras $
 */

#ifndef _SERVER_MODEL_METADATA_H_
#define _SERVER_MODEL_METADATA_H_
 
#include "misc.h"
#include "SizeHistogram.h"

#define MODEL_SIZE_HISTOGRAM_BINS 13	/* 2^13=8192 */

/**
 * The ServerModelMetadata class holds the metadata maintained by the exact
 * caller of the ServerClientConnection::process_command(), which varies per
 * the multithreading model. Its main purpose is to allow keeping unsynchronized
 * statistics information.
 */
class ServerModelMetadata {
private:
public:
	SizeHistogram<MODEL_SIZE_HISTOGRAM_BINS> size_hist;
	
	void recordBufSize(unsigned s) {
		size_hist.addPoint(s);
	}
	
	unsigned getPeakBufSize() {
		return size_hist.getPeakSize();
	}
	
	SizeHistogram<MODEL_SIZE_HISTOGRAM_BINS> *getHistogram() {
		return &size_hist;
	}
	
	void clearHistogram() {
		size_hist.clear();
	}
};

#endif
