/*-
 * This file is a part of the MDCacheD server
 * (c) 2012. Ivan Voras <ivoras@gmail.com>
 * Released under the 2-clause BSDL
 * $Id: ServerMonitorThread.cpp 452 2012-03-20 16:03:25Z ivoras $
 */

#include <err.h>
#include <pthread.h>
#include <signal.h>
#include <assert.h>

#include "queue.h"
#include <list>
#include <iterator>

#include "logger.h"
#include "debug.h"
#include "misc.h"
#include "server.h"
#include "Thread.h"
#include "ServerMonitorThread.h"
#include "ServerModelMetadata.h"

extern time_t current_time;

extern ServerModelMetadata *model_metadata;
extern unsigned int n_model_metadata;

extern unsigned int inbuf_alloc_size;
extern unsigned int outbuf_alloc_size;
extern bool adaptive_reconfiguration;


void
ServerMonitorThread::runAdaptiveReconfiguration()
{
	SizeHistogram<MODEL_SIZE_HISTOGRAM_BINS> totals;
	for (unsigned i = 0; i < n_model_metadata; i++)
		totals.mergeHistogram(model_metadata[i].getHistogram());
	int peak_bin = totals.getPeakBin();
	if (peak_bin == -1)	/* no data points collected */
		return;
	unsigned peak_size = 1 << peak_bin;
	if (peak_size != inbuf_alloc_size) {
		mc_log(LOG_INFO, "Adaptive buffer size target changing to %u", peak_size);
		inbuf_alloc_size = outbuf_alloc_size = peak_size;
	}
	for (unsigned i = 0; i < n_model_metadata; i++)
		model_metadata[i].clearHistogram();
}

void
ServerMonitorThread::run()
{
	while (!stop_requested) {
		if (adaptive_reconfiguration)
			runAdaptiveReconfiguration();
		sleep(MONITOR_THREAD_INTERVAL);
	}
}
