/*-
 * This file is a part of the MDCacheD server
 * (c) 2007.-2012. Ivan Voras <ivoras@gmail.com>
 * Released under the 2-clause BSDL
 * $Id: ServerReaperThread.cpp 484 2012-04-18 15:10:38Z ivoras $
 */
#include <err.h>
#include <pthread.h>
#include <signal.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>

#include "tree.h"
#include "mc_protocol.h"
#include "server.h"
#include "Thread.h"
#include "DataStore.h"
#include "common_utils.h"

#include "ServerReaperThread.h"
 
extern uint64_t server_memory_limit;

/**
 * Constructor.
 */
ServerReaperThread::ServerReaperThread(DataStore *ds) : Thread()
{
	this->ds = ds;
}


/**
 * The reaper thread - does book-keeping on the DataStore.
 */
void
ServerReaperThread::run()
{
	unsigned reaped;
	double t1;
	
	while (!stop_requested) {
		t1 = now();
		reaped = ds->reap_dead_items();
		if (reaped != 0)
			mc_log(LOG_DEBUG, "Dead records reaper reaped %u records in %0.5fs", reaped, now()-t1);

		if (server_memory_limit != 0) {
			t1 = now();
			reaped = ds->reap_free_memory(server_memory_limit);
			if (reaped != 0)
				mc_log(LOG_DEBUG, "Memory limit reaper reaped %u records in %0.5fs", reaped, now()-t1);
		}
		sleep(REAPER_THREAD_INTERVAL);
	}
}

