/*-
 * This file is a part of the MDCacheD server
 * (c) 2007.-2010. Ivan Voras <ivoras@gmail.com>
 * Released under the 2-clause BSDL
 * $Id: ServerWorkerThread.cpp 449 2012-03-20 14:39:42Z ivoras $
 */

#include <sys/param.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <pthread.h>
#include <signal.h>
#include <syslog.h>
#include <assert.h>
#include <stdlib.h>
#include <pthread.h>
#include <err.h>

#include "queue.h"
#include <list>
#include <iterator>

#include "logger.h"
#include "debug.h"
#include "misc.h"
#include "server.h"
#include "Thread.h"
#include "ServerWorkerThread.h"
#include "ServerModelMetadata.h"
#include "ServerClientConnection.h"

extern time_t current_time;

extern ServerModelMetadata *model_metadata;
extern unsigned int n_model_metadata;


ServerWorkerThread::ServerWorkerThread() : Thread()
{
	CRIT_ASSERT(pthread_mutex_init(&queue_mtx, NULL) == 0);
	CRIT_ASSERT(pthread_cond_init(&queue_cv, NULL) == 0);
}


void
ServerWorkerThread::run()
{
	ServerClientConnection **dequeued_conns;
	ServerClientConnection *conn;
	int n_dequeued_conns;
	int i;

	mc_log(LOG_INFO, "Worker thread %p starting", this);

	dequeued_conns = (ServerClientConnection **) malloc(MAX_DEQUEUED_JOBS * sizeof *dequeued_conns);
	if (dequeued_conns == NULL)
		mc_log(LOG_CRIT, "Cannot allocate %d bytes for thread dequeue array", MAX_DEQUEUED_JOBS * sizeof *dequeued_conns);

	while (!stop_requested) {
		DEBUG_STATUS_POINT(st, 'm');
		CRIT_ASSERT(pthread_mutex_lock(&queue_mtx) == 0);
		DEBUG_STATUS_POINT(st, 'c');
		while (queue.empty()) {
			struct timespec ts = {current_time+2, 0};
			int twr = pthread_cond_timedwait(&queue_cv, &queue_mtx, &ts);
			if (twr == 0)
				break;
			else if (twr == ETIMEDOUT) {
				if (stop_requested)
					goto endit;
				continue;
			} else
				mc_log(LOG_CRIT, "pthread_cond_timedwait() failed");
		}

		/* Dequeue up to MAX_DEQUEUED_JOBS */
		n_dequeued_conns = 0;
		while (n_dequeued_conns <= MAX_DEQUEUED_JOBS-1 && !queue.empty()) {
			DEBUG_STATUS_POINT(st, 'q');
			conn = queue.front();
			queue.pop_front();
			queue_len--;
			stat_n_reqs++;
			assert(conn != NULL);
			dequeued_conns[n_dequeued_conns++] = conn;
		}

		CRIT_ASSERT(pthread_mutex_unlock(&queue_mtx) == 0);
		DEBUG_STATUS_POINT(st, 'j');

		/* Process all dequeued connections at once, outside the locks */
		for (i = 0; i < n_dequeued_conns; i++)
			dequeued_conns[i]->process_command(&model_metadata[index]);
	}
endit:
	mc_log(LOG_INFO, "Worker thread %p exiting", this);
}

void
ServerWorkerThread::signalQueueReady() {
	CRIT_ASSERT(pthread_cond_signal(&queue_cv) == 0);
}
