/*-
 * This file is a part of the MDCacheD server
 * (c) 2007.-2011. Ivan Voras <ivoras@gmail.com>
 * Released under the 2-clause BSDL.
 * $Id: ServerWorkerThread.h 452 2012-03-20 16:03:25Z ivoras $
 */

#ifndef _SERVER_WORKER_THREAD_H_
#define _SERVER_WORKER_THREAD_H_

class ServerNetworkThread;

class ServerClientConnection;

class ServerWorkerThread : public Thread {
public:
	unsigned int		index;
	unsigned int		stat_n_reqs;
#ifndef NDEBUG
	struct debug_status_record st;
#endif

	ServerWorkerThread();
	void signalQueueReady();
	virtual void run();

	void queueLock() {
		CRIT_ASSERT(pthread_mutex_lock(&queue_mtx) == 0);
	}

	int queueTryLock() {
		return pthread_mutex_trylock(&queue_mtx);
	}

	void queueUnlock() {
		CRIT_ASSERT(pthread_mutex_unlock(&queue_mtx) == 0);
	}

	void queueSignal() {
		CRIT_ASSERT(pthread_cond_signal(&queue_cv) == 0);
	}

	void queueAdd(ServerClientConnection *scn) {
		queue.push_back(scn);
		queue_len++;
	}

	void setQueueLocker(ServerNetworkThread *nthr) {
		queue_locker = nthr;
	}

	ServerNetworkThread *getQueueLocker() {
		return queue_locker;
	}

private:
	pthread_mutex_t		queue_mtx;
	pthread_cond_t		queue_cv;
	ServerNetworkThread	*queue_locker;	/* if != NULL, a network thread locks the queue */
	std::list<ServerClientConnection*> queue;
	signed int		queue_len;	/* it's signed for debugging purposes */
};

#endif
