/*
  +----------------------------------------------------------------------+
  | MDCacheD client library                                              |
  +----------------------------------------------------------------------+
  | Copyright (c) 2008-2011 Ivan Voras                                        |
  +----------------------------------------------------------------------+
  | This source file is subject to version 3.01 of the PHP license,      |
  | that is bundled with this package in the file LICENSE, and is        |
  | available through the world-wide-web at the following url:           |
  | http://www.php.net/license/3_01.txt                                  |
  | If you did not receive a copy of the PHP license and are unable to   |
  | obtain it through the world-wide-web, please send a note to          |
  | license@php.net so we can mail you a copy immediately.               |
  +----------------------------------------------------------------------+
  | Author: Ivan Voras                                                   |
  +----------------------------------------------------------------------+
*/

/* $Id: mdcached.c 460 2012-03-21 12:44:36Z ivoras $ */

/*
 * Some conventions:
 * 		If a record is not found, NULL is returned
 * 		If there is an error, FALSE is returned
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <assert.h>

#include "php.h"
#include "php_ini.h"
#include "ext/standard/info.h"
#include "php_mdcached.h"

#include <mc_protocol.h>
#include <mc_client.h>


/* If you declare any globals in php_mdcached.h uncomment this:
ZEND_DECLARE_MODULE_GLOBALS(mdcached)
*/

/* True global resources - no need for thread safety here */
static int le_mdcached_resource;

#define le_mdcached_resource_name "Bullet Cache connection resource"

void mc_rsrc_destructor(zend_rsrc_list_entry *rsrc TSRMLS_DC);

/* List of active connections to be closed on request completion (exit) */
/* XXX: make it non-stupid */
#define MAX_CONNS 8
struct mc_connection* active_conns[MAX_CONNS] = {NULL};


/* {{{ mdcached_functions[]
 *
 * Every user visible function must have an entry in mdcached_functions[].
 */
#undef mc_put
#undef mc_get
#undef mc_del
const zend_function_entry mdcached_functions[] = {

	PHP_FE(mc_connect_tcp,			NULL)
	PHP_FE(mc_connect_local,		NULL)
	PHP_FE(mc_result_to_string,		NULL)
	PHP_FE(mc_close_connection,		NULL)
	PHP_FE(mc_perform_handshake,		NULL)
	PHP_FE(mc_put_simple,			NULL)
	PHP_FE(mc_get_simple,			NULL)
	PHP_FE(mc_del_simple,			NULL)
	PHP_FE(mc_get_all_keys,			NULL)
	PHP_FE(mc_put_simple_tags,		NULL)
	PHP_FE(mc_get_simple_tags,		NULL)
	PHP_FE(mc_get_by_tag_values,		NULL)
	PHP_FE(mc_del_by_tag_values,		NULL)
	PHP_FE(mc_atomic_cmpset,		NULL)
	PHP_FE(mc_atomic_add,			NULL)
	PHP_FE(mc_atomic_fetchadd,		NULL)
	PHP_FE(mc_atomic_readandclear,		NULL)
	PHP_FE(mc_atomic_put,			NULL)
	PHP_FE(mc_atomic_get,			NULL)
	PHP_FE(mc_mput,				NULL)
	PHP_FE(mc_mget,				NULL)
	PHP_FE(mc_tstack_push,			NULL)
	PHP_FE(mc_tstack_pop,			NULL)
	PHP_FE(mc_set_tags,			NULL)

	PHP_FALIAS(mc_put, mc_put_simple_tags,	NULL)
	PHP_FALIAS(mc_get, mc_get_simple_tags,	NULL)
	PHP_FALIAS(mc_del, mc_del_simple,	NULL)

	{NULL, NULL, NULL}	/* Must be the last line in mdcached_functions[] */
};
/* }}} */

/*
 * Note: return convention: errors return FALSE, missing records return NULL.
 */

/* {{{ mdcached_module_entry
 */
zend_module_entry mdcached_module_entry = {
#if ZEND_MODULE_API_NO >= 20010901
	STANDARD_MODULE_HEADER,
#endif
	"mdcached",
	mdcached_functions,
	PHP_MINIT(mdcached),
	PHP_MSHUTDOWN(mdcached),
	PHP_RINIT(mdcached),		/* Replace with NULL if there's nothing to do at request start */
	PHP_RSHUTDOWN(mdcached),	/* Replace with NULL if there's nothing to do at request end */
	PHP_MINFO(mdcached),
#if ZEND_MODULE_API_NO >= 20010901
	"1.0.3", /* Replace with version number for your extension */
#endif
	STANDARD_MODULE_PROPERTIES
};
/* }}} */

#ifdef COMPILE_DL_MDCACHED
ZEND_GET_MODULE(mdcached)
#endif

/* {{{ PHP_INI
 */
/* Remove comments and fill if you need to have entries in php.ini
PHP_INI_BEGIN()
    STD_PHP_INI_ENTRY("mdcached.global_value",      "42", PHP_INI_ALL, OnUpdateLong, global_value, zend_mdcached_globals, mdcached_globals)
    STD_PHP_INI_ENTRY("mdcached.global_string", "foobar", PHP_INI_ALL, OnUpdateString, global_string, zend_mdcached_globals, mdcached_globals)
PHP_INI_END()
*/
/* }}} */

/* {{{ php_mdcached_init_globals
 */
/* Uncomment this function if you have INI entries
static void php_mdcached_init_globals(zend_mdcached_globals *mdcached_globals)
{
	mdcached_globals->global_value = 0;
	mdcached_globals->global_string = NULL;
}
*/
/* }}} */

/* {{{ PHP_MINIT_FUNCTION
 */
PHP_MINIT_FUNCTION(mdcached)
{
	/* If you have INI entries, uncomment these lines
	REGISTER_INI_ENTRIES();
	*/
	le_mdcached_resource = zend_register_list_destructors_ex(NULL, mc_rsrc_destructor, le_mdcached_resource_name, module_number);

	REGISTER_LONG_CONSTANT("MC_DEFAULT_INET_PORT", MC_DEFAULT_INET_PORT, CONST_CS | CONST_PERSISTENT);
	REGISTER_STRING_CONSTANT("MC_DEFAULT_UNIX_SOCKET", MC_DEFAULT_UNIX_SOCKET, CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("MCMD_FLAG_ATOMICALL", MCMD_FLAG_ATOMICALL, CONST_CS | CONST_PERSISTENT);

	REGISTER_LONG_CONSTANT("MC_RESULT_OK", 		MC_RESULT_OK, 		CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("MC_RESULT_STATUS", 	MC_RESULT_STATUS, 	CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("MC_RESULT_NETWORK", 	MC_RESULT_NETWORK, 	CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("MC_RESULT_NOT_FOUND", 	MC_RESULT_NOT_FOUND, 	CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("MC_RESULT_UNEXPECTED", 	MC_RESULT_UNEXPECTED, 	CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("MC_INVALID_PARAM", 	MC_INVALID_PARAM, 	CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("MC_RESULT_STATE", 	MC_RESULT_STATE, 	CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("MC_ENOMEM", 		MC_ENOMEM, 		CONST_CS | CONST_PERSISTENT);

	return SUCCESS;
}
/* }}} */

/* {{{ PHP_MSHUTDOWN_FUNCTION
 */
PHP_MSHUTDOWN_FUNCTION(mdcached)
{
	/* uncomment this line if you have INI entries
	UNREGISTER_INI_ENTRIES();
	*/
	return SUCCESS;
}
/* }}} */

/* Remove if there's nothing to do at request start */
/* {{{ PHP_RINIT_FUNCTION
 */
PHP_RINIT_FUNCTION(mdcached)
{
	return SUCCESS;
}
/* }}} */

/* Close active connections on request completion. */
/* {{{ PHP_RSHUTDOWN_FUNCTION
 */
PHP_RSHUTDOWN_FUNCTION(mdcached)
{
	int i;
	
	for (i = 0; i < MAX_CONNS; i++)
		if (active_conns[i] != NULL) {
			mc_close_connection(active_conns[i]);
			active_conns[i] = NULL;
		}
	return SUCCESS;
}
/* }}} */

/* {{{ PHP_MINFO_FUNCTION
 */
PHP_MINFO_FUNCTION(mdcached)
{
	char version[5], port[5];
	sprintf(version, "%d", MC_PROTOCOL_VERSION);
	sprintf(port, "%d", MC_DEFAULT_INET_PORT);

	php_info_print_table_start();
	php_info_print_table_header(2, "Bullet cache support", "enabled");
	php_info_print_table_row(2, "PHP module versioning Id", "$Id: mdcached.c 460 2012-03-21 12:44:36Z ivoras $");
	php_info_print_table_row(2, "Protocol header versioning Id", MC_PROTOCOL_VERSION_ID);
	php_info_print_table_row(2, "Client header versioning Id", MC_CLIENT_VERSION_ID);
	php_info_print_table_row(2, "Protocol version", version);
#ifdef TAGS_64BIT
	php_info_print_table_row(2, "64-bit tags", "enabled");
#endif
	php_info_print_table_row(2, "Default TCP port", port);
	php_info_print_table_row(2, "Default Unix socket", MC_DEFAULT_UNIX_SOCKET);
	php_info_print_table_end();

	/* Remove comments if you have entries in php.ini
	DISPLAY_INI_ENTRIES();
	*/
}
/* }}} */


/**
 * Parses the given record / data entry into a PHP associative array with record
 * key and value, and with a special additional entry with the key '_tags' which
 * contains the record tags.
 */
static inline zval*
_php_mc_parse_single_record(struct mc_data_entry *mce)
{
	zval *result;
	zval *tags;
	struct mc_tag *rtags;
	int i;
	
	ALLOC_INIT_ZVAL(result);
	array_init(result);
	add_assoc_stringl_ex(result, mc_data_entry_key(mce), mce->name_size + 1, 
		mc_data_entry_value(mce), mce->data_size, 1);
	
	rtags = mc_data_entry_tags(mce);
	ALLOC_INIT_ZVAL(tags);
	array_init(tags);
	for (i = 0; i < mce->n_tags; i++)
		add_index_long(tags, rtags[i].key, rtags[i].value);
	
	add_assoc_zval(result, "_tags", tags);
	return result;
}


/**
 * Composes an associative array with three elements: 
 * the first one has the key "_key" and the value of the given key
 * the second has the key "_value" and the value of the given data
 * the third has the key "_tags" and it is another array containing the given tags.
 */
static inline zval*
_php_mc_compose_record_entry(char *key, unsigned key_size, char *data, unsigned data_size, struct mc_tag *tags, unsigned n_tags)
{
	zval *zresult, *ztags;
	int i;
	
	ALLOC_INIT_ZVAL(zresult);
	array_init(zresult);
	add_assoc_stringl(zresult, "_key", key, key_size, 1);
	add_assoc_stringl(zresult, "_value", data, data_size, 1);
	
	ALLOC_INIT_ZVAL(ztags);
	array_init(ztags);
	for (i = 0; i < n_tags; i++)
		add_index_long(ztags, tags[i].key, tags[i].value);
	add_assoc_zval(zresult, "_tags", ztags);
	
	return zresult;
}


/**
 * Parses the given struct mc_multidata_result into a PHP associative array and
 * returns this array as a zval.
 *
 * The associative array contains one element for each record in the multidata
 * structure. The element are keyed with the record keys (as strings), and the
 * element's values are nested associative arrays, which contain three special
 * elements with the keys "_key", "_value" and "_tags", as described for the
 * _php_mc_compose_record_entry() function.
 */
static inline zval*
_php_mc_parse_multidata(struct mc_multidata_result *mdres)
{
	zval *result;
	unsigned i;

	ALLOC_INIT_ZVAL(result);
	array_init(result);

	for (i = 0; i < mdres->n_records; i++) {
		struct mc_data_entry *rp;

		rp = mdres->records[i];
		assert(mc_data_entry_key(rp) != NULL);
		assert(mc_data_entry(rp) != NULL);
		add_assoc_zval_ex(result, mc_data_entry_key(rp), rp->name_size + 1,
			_php_mc_compose_record_entry(
				mc_data_entry_key(rp), mc_data_entry_key_size(rp),
				mc_data_entry_value(rp), mc_data_entry_value_size(rp),
				mc_data_entry_tags(rp), mc_data_entry_n_tags(rp)));
	}
	return result;
}


/**
 * Parses the given struct mc_multistring_result into a PHP array, and returns
 * this array as a zval.
 */
static inline zval*
_php_mc_parse_multistring(struct mc_multistring_result *msres)
{
	zval *result;
	unsigned i;

	ALLOC_INIT_ZVAL(result);
	array_init(result);

	for (i = 0; i < msres->n_records; i++)
		add_next_index_stringl(result, msres->records[i]->data, msres->records[i]->str_len, 1);

	return result;
}


/**
 * Parse the zval which should be a tags array (i.e. array(tag=>value)).
 */
static int
_php_parse_tags_from_zval(zval *ztags, struct mc_tag **tags, unsigned *n_tags)
{
	int i;
	zval **data;
	HashTable *ztags_hash;
	HashPosition hpos;

	if (ztags == NULL) {
		*n_tags = 0;
		*tags = NULL;
		return FALSE;
	}
	
	ztags_hash = Z_ARRVAL_P(ztags);
	*n_tags = zend_hash_num_elements(ztags_hash);
	*tags = malloc(*n_tags * sizeof(struct mc_tag));
	i = 0;
	for (zend_hash_internal_pointer_reset_ex(ztags_hash, &hpos); zend_hash_get_current_data_ex(ztags_hash, (void**) &data, &hpos) == SUCCESS; zend_hash_move_forward_ex(ztags_hash, &hpos)) {
		long index;

		if (zend_hash_get_current_key_ex(ztags_hash, NULL, NULL, &index, 0, &hpos) == HASH_KEY_IS_LONG) {
			tag_key_t tkey = index;
			tag_value_t tval;

			if (Z_TYPE_PP(data) != IS_LONG) {
				php_error(E_ERROR, "Array keys and values for tags must be integers");
				free(*tags);
				*n_tags = 0;
				*tags = NULL;
				return FALSE;
			}
			tval = Z_LVAL_PP(data);

			(*tags)[i].key = tkey;
			(*tags)[i].value = tval;
			i++;
		} else {
			php_error(E_ERROR, "Array keys and values for tags must be integers");
			free(*tags);
			*n_tags = 0;
			*tags = NULL;
			return FALSE;
		}
	}
	return TRUE;
}


/**
 * Resource destructor for mc_connection resource.
 * @param	rsrc
 */
void mc_rsrc_destructor(zend_rsrc_list_entry *rsrc TSRMLS_DC)
{
	struct mc_connection *mc = rsrc->ptr;

	if (mc != NULL)
		mc_close_connection(mc);
}


/**
 * Returns result (error) message string from integer result code for mdcached.
 * @param	int		code
 * @return	string
 */
PHP_FUNCTION(mc_result_to_string)
{
	long code;
	char *str;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "l", &code) == FAILURE)
		RETURN_NULL();

	str = mc_result_to_string(code);
	RETURN_STRING(str, 1);
}


/**
 * Connects to a remote mdcached server via TCP.
 * @param	string		host_name
 * @param	int		port		MC_DEFAULT_TCP_PORT
 * @param	bool		do_handshake	true
 * @return	resource
 */
PHP_FUNCTION(mc_connect_tcp)
{
	char *host_name;
	long port = MC_DEFAULT_INET_PORT;
	zend_bool do_handshake = 1;
	struct mc_connection *mc;
	char *err = NULL;
	int i;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|lb", &host_name, &port, &do_handshake) == FAILURE)
		RETURN_NULL();

	mc = mc_connect_tcp(host_name, port, do_handshake, &err);

	if (err != NULL || mc == NULL) {
		php_error(E_ERROR, "Error in mc_connect_tcp: %s", err);
		RETURN_FALSE;
	}
	
	for (i = 0; i < MAX_CONNS; i++)
		if (active_conns[i] == NULL) {
			active_conns[i] = mc;
			break;
		}

	ZEND_REGISTER_RESOURCE(return_value, mc, le_mdcached_resource);
}


/**
 * Connects to a local mdcached server via Unix sockets.
 * @param	string		unix_socket
 * @param	bool		do_handshake	true
 * @return	resource
 */
PHP_FUNCTION(mc_connect_local)
{
	char *unix_socket;
	zend_bool do_handshake = 1;
	struct mc_connection *mc;
	char *err = NULL;
	int i;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|b", &unix_socket, &do_handshake) == FAILURE)
		RETURN_NULL();

	mc = mc_connect_local(unix_socket, do_handshake, &err);

	if (err != NULL) {
		php_error(E_ERROR, "Error in mc_connect_local: %s", err);
		RETURN_FALSE;
		free(err);
	}
	
	for (i = 0; i < MAX_CONNS; i++)
		if (active_conns[i] == NULL) {
			active_conns[i] = mc;
			break;
		}

	ZEND_REGISTER_RESOURCE(return_value, mc, le_mdcached_resource);
}


/**
 * Closes the connection.
 * @param	resource	mc_connection
 * @param	bool
 */
PHP_FUNCTION(mc_close_connection)
{
	struct mc_connection *mc;
	zval *zmc;
	int i;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r", &zmc) == FAILURE)
		RETURN_FALSE;

	ZEND_FETCH_RESOURCE(mc, struct mc_connection*, &zmc, -1, le_mdcached_resource_name, le_mdcached_resource);

	for (i = 0; i < MAX_CONNS; i++)
		if (active_conns[i] == mc) {
			active_conns[i] = NULL;
			break;
		}
	
	mc_close_connection(mc);

	zend_list_delete(Z_LVAL_P(zmc));

	RETURN_TRUE;
}


/**
 * Performs a diagnostic handshake / NOP to determine if the server is alive and
 * there's mutual understanding of the protocol (e.g. endian-wise).
 *
 * @param	resource	mc_connection
 * @return	long
 */
PHP_FUNCTION(mc_perform_handshake)
{
	struct mc_connection *mc;
	zval *zmc;
	char *err = NULL;
	long ret;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r", &zmc) == FAILURE)
		RETURN_FALSE;

	ZEND_FETCH_RESOURCE(mc, struct mc_connection*, &zmc, -1, le_mdcached_resource_name, le_mdcached_resource);

	ret = mc_perform_handshake(mc, &err);

	if (err != NULL) {
		php_error(E_ERROR, "Error in mc_perform_handshake: %s", err);
		free(err);
		RETURN_FALSE;
	}

	RETURN_LONG(ret);
}


/**
 * Performs a simple PUT operation to the mdcached.
 *
 * @param	resource	mc_connection
 * @param	string		key
 * @param	string		value
 * @param	long		expire_time
 * @return	bool
 */
PHP_FUNCTION(mc_put_simple)
{
	struct mc_connection *mc;
	zval *zmc;
	char *err = NULL, *key, *value;
	int key_len, value_len;
	long expire_time = 0, ret;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rss|l", &zmc, &key, &key_len, &value, &value_len, &expire_time) == FAILURE)
		RETURN_FALSE;

	ZEND_FETCH_RESOURCE(mc, struct mc_connection*, &zmc, -1, le_mdcached_resource_name, le_mdcached_resource);

	ret = mc_put_simple(mc, key, key_len, value, value_len, expire_time, &err);
	if (err != NULL) {
		php_error(E_ERROR, "Error in mc_put_simple: %s", err);
		free(err);
		RETURN_FALSE;
	}
	if (ret != MC_RESULT_OK) {
		php_error(E_ERROR, "mc_put_simple() returned an error");
		RETURN_FALSE;
	}

	RETURN_LONG(ret);
}


/**
 * Perform a simple GET operation on a record in mdcached.
 *
 * @param	resource	mc_connection
 * @param	string		key
 * @return	string
 */
PHP_FUNCTION(mc_get_simple)
{
	struct mc_connection *mc;
	zval *zzmc;
	long ret;
	char *key;
	int key_len;
	char *value, *err = NULL;
	unsigned int value_len;

	if (ZEND_NUM_ARGS() != 2)
		WRONG_PARAM_COUNT;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rs", &zzmc, &key, &key_len) == FAILURE)
		RETURN_FALSE;

	ZEND_FETCH_RESOURCE(mc, struct mc_connection*, &zzmc, -1,
		le_mdcached_resource_name, le_mdcached_resource);

	ret = mc_get_simple(mc, key, key_len, &value, &value_len, &err);
	if (ret == MC_RESULT_NOT_FOUND)
		RETURN_NULL();
	if (ret != MC_RESULT_OK || err != NULL) {
		php_error(E_ERROR, "Error in mc_get_simple: %s", err);
		free(err);
		RETURN_FALSE;
	}

	RETVAL_STRINGL(value, value_len, 1);
	free(value);
}


/**
 * Perform a simple DELETE operation on a record in mdcached.
 *
 * @param	resource	mc_connection
 * @param	string		key
 * @return	bool
 */
PHP_FUNCTION(mc_del_simple)
{
	struct mc_connection *mc;
	zval *zzmc;
	char *key, *err = NULL;
	int key_len, ret;
	
	if (ZEND_NUM_ARGS() != 2)
		WRONG_PARAM_COUNT;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rs", &zzmc, &key, &key_len) == FAILURE)
		RETURN_FALSE;
	
	ZEND_FETCH_RESOURCE(mc, struct mc_connection*, &zzmc, -1,
		le_mdcached_resource_name, le_mdcached_resource);

	ret = mc_del_simple(mc, key, key_len, &err);
	if (ret == MC_RESULT_NOT_FOUND)
		RETURN_FALSE;
	if (ret != MC_RESULT_OK || err != NULL) {
		php_error(E_ERROR, "Error in mc_get_simple: %s", err);
		free(err);
		RETURN_FALSE;
	}
	RETURN_TRUE;
}


/**
 * Puts a tagged record into the cache server.
 *
 * @param	resource	mc_connection
 * @param	string		key
 * @param	string		value
 * @param	array		tags
 * @param	long		expire_time
 * @return	bool
 */
PHP_FUNCTION(mc_put_simple_tags)
{
	struct mc_connection *mc;
	zval *zmc, *ztags = NULL;
	char *err = NULL, *key, *value;
	int key_len, value_len;
	unsigned n_tags = 0;
	struct mc_tag *tags;
	long expire_time = 0;
	long ret;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rss|al", &zmc, &key, &key_len, &value, &value_len, &ztags, &expire_time) == FAILURE)
		RETURN_FALSE;

	ZEND_FETCH_RESOURCE(mc, struct mc_connection*, &zmc, -1, le_mdcached_resource_name, le_mdcached_resource);
	_php_parse_tags_from_zval(ztags, &tags, &n_tags);
	ret = mc_put_simple_tags(mc, key, key_len, value, value_len, tags, n_tags, expire_time, &err);
	if (tags != NULL)
		free(tags);

	if (err != NULL) {
		php_error(E_ERROR, "Error in mc_put_simple_tags: %s", err);
		free(err);
		RETURN_FALSE;
	}

	RETURN_LONG(ret);
}


/**
 * Gets a record, including its tags. This function returns an array which looks
 * like array('key' => 'value', '_tags' => array(1 => 2 ...)).
 *
 * @param	resource	mc_connection
 * @param	string		key
 * @return	array
 */
PHP_FUNCTION(mc_get_simple_tags)
{
	struct mc_connection *mc;
	zval *zmc, *zret, *zrec;
	char *key, *data, *err;
	struct mc_tag *tags;
	unsigned key_len, data_len, n_tags;
	int ret;
	
	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rs", &zmc, &key, &key_len) == FAILURE)
		RETURN_FALSE;
	ZEND_FETCH_RESOURCE(mc, struct mc_connection*, &zmc, -1, le_mdcached_resource_name, le_mdcached_resource);
	ret = mc_get_simple_tags(mc, key, key_len, &data, &data_len, &tags, &n_tags, &err);
	if (ret == MC_RESULT_OK) {
		zrec = _php_mc_compose_record_entry(key, key_len, data, data_len, tags, n_tags);
		free(data);
		free(tags);
		
		ALLOC_INIT_ZVAL(zret);
		array_init(zret);
		add_assoc_zval_ex(zret, key, key_len + 1, zrec);
		RETURN_ZVAL(zret, 0, 0);
	} else {
		php_error(E_ERROR, "mc_get_simple_tags: %d,%s", ret, err);
		RETURN_FALSE;
	}
}


/**
 * Retrieve records which have the given tag key and any of the given tag values.
 * The result is a key-value array.
 *
 * @param	resource	mc_connection
 * @param	int		tag_key
 * @param	array(int)	tag_values
 * @return	array
 */
PHP_FUNCTION(mc_get_by_tag_values)
{
	struct mc_connection *mc;
	zval *zmc, *zvals, **data;
	long tag_key;
	HashTable *zvals_hash;
	HashPosition hpos;
	unsigned n_values, i = 0;
	tag_value_t *values;
	struct mc_multidata_result *mdres;
	char *err;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rla", &zmc, &tag_key, &zvals) == FAILURE)
		RETURN_FALSE;

	ZEND_FETCH_RESOURCE(mc, struct mc_connection*, &zmc, -1, le_mdcached_resource_name, le_mdcached_resource);

	zvals_hash = Z_ARRVAL_P(zvals);
	n_values = zend_hash_num_elements(zvals_hash);
	values = malloc(n_values * sizeof(*values));

	for (zend_hash_internal_pointer_reset_ex(zvals_hash, &hpos); zend_hash_get_current_data_ex(zvals_hash, (void**) &data, &hpos) == SUCCESS; zend_hash_move_forward_ex(zvals_hash, &hpos)) {
		long index;

		if (zend_hash_get_current_key_ex(zvals_hash, NULL, NULL, &index, 0, &hpos) == HASH_KEY_NON_EXISTANT) {
			php_error(E_ERROR, "Invalid hash element");
			free(values);
			RETURN_FALSE;
		}
		if (Z_TYPE_PP(data) != IS_LONG) {
			php_error(E_ERROR, "Array values must be integers");
			free(values);
			RETURN_FALSE;
		}
		values[i++] = Z_LVAL_PP(data);
	}

	if (mc_get_by_tag_values(mc, tag_key, values, n_values, &mdres, &err) != MC_RESULT_OK) {
		php_error(E_ERROR, "Failure in mc_get_by_tag_values(): %s", err);
		if (err != NULL)
			free(err);
		free(values);
		RETURN_FALSE;
	}
	zvals = _php_mc_parse_multidata(mdres);
	mc_multidata_result_free(mdres);
	free(values);

	RETURN_ZVAL(zvals, 0, 0);
}


/**
 * Delete records which have the given tag key and any of the given tag values.
 * The result is a key-value array.
 *
 * @param	resource	mc_connection
 * @param	int		tag_key
 * @param	array(int)	tag_values
 * @return	bool
 */
PHP_FUNCTION(mc_del_by_tag_values)
{
	struct mc_connection *mc;
	zval *zmc, *zvals, **data;
	long tag_key;
	HashTable *zvals_hash;
	HashPosition hpos;
	unsigned n_values, i = 0, n_del = 0;
	tag_value_t *values;
	char *err;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rla", &zmc, &tag_key, &zvals) == FAILURE)
		RETURN_FALSE;

	ZEND_FETCH_RESOURCE(mc, struct mc_connection*, &zmc, -1, le_mdcached_resource_name, le_mdcached_resource);

	zvals_hash = Z_ARRVAL_P(zvals);
	n_values = zend_hash_num_elements(zvals_hash);
	values = malloc(n_values * sizeof(*values));

	for (zend_hash_internal_pointer_reset_ex(zvals_hash, &hpos); zend_hash_get_current_data_ex(zvals_hash, (void**) &data, &hpos) == SUCCESS; zend_hash_move_forward_ex(zvals_hash, &hpos)) {
		long index;

		if (zend_hash_get_current_key_ex(zvals_hash, NULL, NULL, &index, 0, &hpos) == HASH_KEY_NON_EXISTANT) {
			php_error(E_ERROR, "Invalid hash element");
			free(values);
			RETURN_FALSE;
		}
		if (Z_TYPE_PP(data) != IS_LONG) {
			php_error(E_ERROR, "Array values must be integers");
			free(values);
			RETURN_FALSE;
		}
		values[i++] = Z_LVAL_PP(data);
	}

	if (mc_del_by_tag_values(mc, tag_key, values, n_values, &n_del, &err) != MC_RESULT_OK) {
		php_error(E_ERROR, "Failure in mc_get_by_tag_values(): %s", err);
		if (err != NULL)
			free(err);
		free(values);
		RETURN_FALSE;
	}
	free(values);

	RETURN_LONG(n_del);
}


/**
 * Retrieve all record keys from the cache server. This is an expensive
 * operation. The flags argument can be MCMD_FLAG_ATOMICALL. The result is a
 * simple array of strings.
 *
 * @param	resource	mc_connection
 * @param	int		flags
 * @return	array
 */
PHP_FUNCTION(mc_get_all_keys)
{
	struct mc_connection *mc;
	zval *zmc, *zkeys;
	long flags = 0;
	char *err;
	struct mc_multistring_result *mr;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r|l", &zmc, &flags) == FAILURE)
		RETURN_FALSE;

	ZEND_FETCH_RESOURCE(mc, struct mc_connection*, &zmc, -1, le_mdcached_resource_name, le_mdcached_resource);

	if (mc_get_all_keys(mc, flags, &mr, &err) != MC_RESULT_OK) {
		php_error(E_ERROR, "Failure in mc_get_all_keys(): %s", err);
		if (err != NULL)
			free(err);
		RETURN_FALSE;
	}
	zkeys = _php_mc_parse_multistring(mr);
	mc_multistring_result_free(mr);

	RETURN_ZVAL(zkeys, 0, 0);
}


/**
 * Perform the CMPSET operation on the specified record. The CMPSET(arg1, arg2)
 * operation is defined as "if (value == arg1) { value = arg2; return 1 } else { return 0 }".
 *
 * @param	resource	mc_connection
 * @param	string		key
 * @param	int		arg1
 * @param	int		arg2
 * @return	bool
 */
PHP_FUNCTION(mc_atomic_cmpset)
{
	struct mc_connection *mc;
	zval *zmc;
	int64_t arg1, arg2;
	long larg1, larg2;
	char *key, *err;
	int key_len, res;
	int64_t op_res;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rsll", &zmc, &key, &key_len, &larg1, &larg2) == FAILURE)
		RETURN_FALSE;
	arg1 = larg1;
	arg2 = larg2;
	ZEND_FETCH_RESOURCE(mc, struct mc_connection*, &zmc, -1, le_mdcached_resource_name, le_mdcached_resource);
	if (mc_atomic_cmpset(mc, key, key_len, arg1, arg2, &op_res, &err) != MC_RESULT_OK) {
		php_error(E_ERROR, "Failure in mc_atomic_cmpset(): %s", err);
		if (err != NULL)
			free(err);
		RETURN_NULL();
	}
	RETURN_BOOL(op_res);
}


/**
 * Perform the ADD operation on the specified record. The ADD(arg1)
 * operation is defined as "value += arg; return value".
 *
 * @param	resource	mc_connection
 * @param	string		key
 * @param	int		arg
 * @return	int
 */
PHP_FUNCTION(mc_atomic_add)
{
	struct mc_connection *mc;
	zval *zmc;
	int64_t arg1;
	long larg1, res;
	char *key, *err;
	int key_len;
	int64_t op_res;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rsl", &zmc, &key, &key_len, &larg1) == FAILURE)
		RETURN_FALSE;
	arg1 = larg1;
	ZEND_FETCH_RESOURCE(mc, struct mc_connection*, &zmc, -1, le_mdcached_resource_name, le_mdcached_resource);
	if (mc_atomic_add(mc, key, key_len, arg1, &op_res, &err) != MC_RESULT_OK) {
		php_error(E_ERROR, "Failure in mc_atomic_cmpset(): %s", err);
		if (err != NULL)
			free(err);
		RETURN_FALSE;
	}
	RETURN_LONG(op_res);
}


/**
 * Perform the FETCHADD operation on the specified record. The FETCHADD(arg1)
 * operation is defined as "tmp = value; value += arg; return tmp".
 *
 * @param	resource	mc_connection
 * @param	string		key
 * @param	int		arg
 * @return	int
 */
PHP_FUNCTION(mc_atomic_fetchadd)
{
	struct mc_connection *mc;
	zval *zmc;
	int64_t arg1;
	long larg1, res;
	char *key, *err;
	int key_len;
	int64_t op_res;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rsl", &zmc, &key, &key_len, &larg1) == FAILURE)
		RETURN_FALSE;
	arg1 = larg1;
	ZEND_FETCH_RESOURCE(mc, struct mc_connection*, &zmc, -1, le_mdcached_resource_name, le_mdcached_resource);
	if (mc_atomic_fetchadd(mc, key, key_len, arg1, &op_res, &err) != MC_RESULT_OK) {
		php_error(E_ERROR, "Failure in mc_atomic_cmpset(): %s", err);
		if (err != NULL)
			free(err);
		RETURN_FALSE;
	}
	RETURN_LONG(op_res);
}


/**
 * Perform the READANDCLEAR operation on the specified record. The READANDCLEAR()
 * operation is defined as "tmp = value; value = 0; return tmp".
 *
 * @param	resource	mc_connection
 * @param	string		key
 * @return	int
 */
PHP_FUNCTION(mc_atomic_readandclear)
{
	struct mc_connection *mc;
	zval *zmc;
	char *key, *err;
	long res;
	int key_len;
	int64_t op_res;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rs", &zmc, &key, &key_len) == FAILURE)
		RETURN_FALSE;
	ZEND_FETCH_RESOURCE(mc, struct mc_connection*, &zmc, -1, le_mdcached_resource_name, le_mdcached_resource);
	if (mc_atomic_readandclear(mc, key, key_len, &op_res, &err) != MC_RESULT_OK) {
		php_error(E_ERROR, "Failure in mc_atomic_cmpset(): %s", err);
		if (err != NULL)
			free(err);
		RETURN_FALSE;
	}
	RETURN_LONG(res);
}


/**
 * @param 	resource	mc_connection
 * @param	string		key
 * @param	long		value
 * @param	array		tags
 * @return	int		MC_RESULT
 */
PHP_FUNCTION(mc_atomic_put)
{
	struct mc_connection *mc;
	zval *zmc, *ztags = NULL;
	int64_t val;
	char *key, *err = NULL;
	long lval, expire_time = 0;
	struct mc_tag *tags = NULL;
	unsigned n_tags;
	int key_len, ret;
	
	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rsl|al", &zmc, &key, &key_len, &lval, &ztags, &expire_time) == FAILURE)
		RETURN_FALSE;
	ZEND_FETCH_RESOURCE(mc, struct mc_connection*, &zmc, -1, le_mdcached_resource_name, le_mdcached_resource);
	val = lval;
	_php_parse_tags_from_zval(ztags, &tags, &n_tags);
	ret = mc_atomic_put(mc, key, key_len, val, tags, n_tags, expire_time, &err);
	if (tags != NULL)
		free(tags);
	if (err != NULL) {
		php_error(E_ERROR, "Error in mc_atomic_put: %s", err);
		free(err);
		RETURN_FALSE;
	}
	RETURN_LONG(ret);
}


/**
 * @param	resource	mc_connection
 * @param	string		key
 * @return	long
 */
PHP_FUNCTION(mc_atomic_get)
{
	struct mc_connection *mc;
	zval *zmc;
	char *key, *err;
	int64_t val = -1;
	int key_len;
	
	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rs", &zmc, &key, &key_len) == FAILURE)
		RETURN_FALSE;
	ZEND_FETCH_RESOURCE(mc, struct mc_connection*, &zmc, -1, le_mdcached_resource_name, le_mdcached_resource);
	if (mc_atomic_get(mc, key, key_len, &val, &err) == MC_RESULT_OK)
		RETURN_LONG(val);
	if (err != NULL)
		php_error(E_ERROR, "Error in mc_atomic_get: %s", err);
	RETURN_FALSE;
}


/**
 * Put multiple key-value records into the cache server.
 *
 * @param	resource	mc_connection
 * @param	records		array
 * @param	int		exptime
 * @param	int		flags
 * @param	array		tags
 */
PHP_FUNCTION(mc_mput)
{
	struct mc_connection *mc;
	zval *zmc, *zrecords, *ztags = NULL, **data;
	HashTable *zhash;
	HashPosition hpos;
	struct mc_tag *tags = NULL;
	unsigned i;
	int n_records, n_tags = 0, res;
	long exptime = 0, flags = 0;
	struct mc_data_entry **records;
	char *err;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ra|lla", &zmc, &zrecords, &exptime, &flags, &ztags) == FAILURE)
		RETURN_FALSE;
	ZEND_FETCH_RESOURCE(mc, struct mc_connection*, &zmc, -1, le_mdcached_resource_name, le_mdcached_resource);

	if (ztags != NULL) {
		zhash = Z_ARRVAL_P(ztags);
		n_tags = zend_hash_num_elements(zhash);
		tags = malloc(n_tags * sizeof(struct mc_tag));

		for (zend_hash_internal_pointer_reset_ex(zhash, &hpos), i = 0; zend_hash_get_current_data_ex(zhash, (void**) &data, &hpos) == SUCCESS; zend_hash_move_forward_ex(zhash, &hpos)) {
			long index;
			tag_key_t tkey;
			tag_value_t tvalue;

			if (zend_hash_get_current_key_ex(zhash, NULL, 0, &index, 0, &hpos) == HASH_KEY_IS_LONG) {
				if (Z_TYPE_PP(data) != IS_LONG) {
					int j;
					php_error(E_ERROR, "Array keys and values for tags must be integers");
					free(tags);
					RETURN_FALSE;
				}

				tkey = index;
				tvalue = Z_LVAL_PP(data);
				tags[i].key = tkey;
				tags[i].value = tvalue;
				i++;
			} else
				php_error(E_ERROR, "Array keys and values for tags must be integers");
		}
	}

	zhash = Z_ARRVAL_P(zrecords);
	n_records = zend_hash_num_elements(zhash);
	records = malloc(n_records * sizeof(*records));
	for (zend_hash_internal_pointer_reset_ex(zhash, &hpos), i = 0; zend_hash_get_current_data_ex(zhash, (void**) &data, &hpos) == SUCCESS; zend_hash_move_forward_ex(zhash, &hpos)) {
		long index;
		char *key;
		int key_len;
		char *value;
		unsigned value_len;

		if (zend_hash_get_current_key_ex(zhash, &key, &key_len, &index, 0, &hpos) == HASH_KEY_IS_STRING) {
			key_len--;
			if (Z_TYPE_PP(data) != IS_STRING) {
				int j;
				php_error(E_ERROR, "Array keys and values for new records must be strings");
				for (j = 0; j < i; j++)
					free(records[j]);
				free(records);
				RETURN_FALSE;
			}

			value = Z_STRVAL_PP(data);
			value_len = Z_STRLEN_PP(data);
			mc_build_data_entry(&records[i], key, key_len, value, value_len, tags, n_tags, exptime, &mc_client_seq);
			i++;
		} else
			php_error(E_ERROR, "Array keys and values for new records must be strings");
	}
	res = mc_mput(mc, n_records, records, flags, &err);
	for (i = 0; i < n_records; i++)
		free(records[i]);
	free(records);
	if (res != MC_RESULT_OK) {
		php_error(E_ERROR, "Error in mc_mput(): %s", err);
		if (err != NULL)
			free(err);
		RETURN_FALSE;
	}
	RETURN_TRUE;
}


/**
 * Retrieve multiple named (by key) records from the cache server.
 *
 * @param 	mc		mc_connection
 * @param	keys		array
 * @param	flags		int
 * @return	array
 */
PHP_FUNCTION(mc_mget)
{
	struct mc_connection *mc;
	zval *zmc, *zkeys, *zres, **data;
	HashTable *zhash;
	HashPosition hpos;
	char *err, **keys;
	uint16_t *key_lengths;
	struct mc_multidata_result *mr;
	unsigned n_keys, i;
	long flags = 0;
	int res;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ra|l", &zmc, &zkeys, &flags) == FAILURE)
		RETURN_FALSE;
	ZEND_FETCH_RESOURCE(mc, struct mc_connection*, &zmc, -1, le_mdcached_resource_name, le_mdcached_resource);

	zhash = Z_ARRVAL_P(zkeys);
	n_keys = zend_hash_num_elements(zhash);
	keys = malloc(n_keys * sizeof(*keys));
	key_lengths = malloc(n_keys * sizeof(*key_lengths));
	for (zend_hash_internal_pointer_reset_ex(zhash, &hpos), i = 0; zend_hash_get_current_data_ex(zhash, (void**) &data, &hpos) == SUCCESS; zend_hash_move_forward_ex(zhash, &hpos)) {
		long index;
		char *key;
		int key_len;

		if (zend_hash_get_current_key_ex(zhash, NULL, NULL, &index, 0, &hpos) == HASH_KEY_NON_EXISTANT) {
			php_error(E_ERROR, "Error retrieving the array of keys");
			free(keys);
			free(key_lengths);
			RETURN_FALSE;
		}
		if (Z_TYPE_PP(data) != IS_STRING) {
			php_error(E_ERROR, "Keys array must only contain strings");
			free(keys);
			free(key_lengths);
			RETURN_FALSE;
		}
		key = Z_STRVAL_PP(data);
		key_len = Z_STRLEN_PP(data);
		keys[i] = key;
		key_lengths[i] = key_len;
		i++;
	}
	res = mc_mget(mc, n_keys, key_lengths, keys, flags, &mr, &err);
	free(keys);
	free(key_lengths);
	if (res != MC_RESULT_OK) {
		php_error(E_ERROR, "Error in mc_mget(): %s", err);
		if (err != NULL)
			free(err);
		RETURN_FALSE;
	}
	zres = _php_mc_parse_multidata(mr);
	RETURN_ZVAL(zres, 0, 0);
}


/**
 * Delete multiple named (by key) records from the cache server. Returns the
 * number of records deleted.
 *
 * @param 	mc		mc_connection
 * @param	keys		array
 * @param	flags		int
 * @return	int
 */
PHP_FUNCTION(mc_mdel)
{
	struct mc_connection *mc;
	zval *zmc, *zkeys, *zres, **data;
	HashTable *zhash;
	HashPosition hpos;
	char *err, **keys;
	uint16_t *key_lengths;
	unsigned n_keys, i, n_del = 0;
	long flags = 0;
	int res;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ra|l", &zmc, &zkeys, &flags) == FAILURE)
		RETURN_FALSE;
	ZEND_FETCH_RESOURCE(mc, struct mc_connection*, &zmc, -1, le_mdcached_resource_name, le_mdcached_resource);

	zhash = Z_ARRVAL_P(zkeys);
	n_keys = zend_hash_num_elements(zhash);
	keys = malloc(n_keys * sizeof(*keys));
	key_lengths = malloc(n_keys * sizeof(*key_lengths));
	for (zend_hash_internal_pointer_reset_ex(zhash, &hpos), i = 0; zend_hash_get_current_data_ex(zhash, (void**) &data, &hpos) == SUCCESS; zend_hash_move_forward_ex(zhash, &hpos)) {
		long index;
		char *key;
		int key_len;

		if (zend_hash_get_current_key_ex(zhash, NULL, NULL, &index, 0, &hpos) == HASH_KEY_NON_EXISTANT) {
			php_error(E_ERROR, "Error retrieving the array of keys");
			free(keys);
			free(key_lengths);
			RETURN_FALSE;
		}
		if (Z_TYPE_PP(data) != IS_STRING) {
			php_error(E_ERROR, "Keys array must only contain strings");
			free(keys);
			free(key_lengths);
			RETURN_FALSE;
		}
		key = Z_STRVAL_PP(data);
		key_len = Z_STRLEN_PP(data);
		keys[i] = key;
		key_lengths[i] = key_len;
		i++;
	}
	res = mc_mdel(mc, n_keys, key_lengths, keys, flags, &n_del, &err);
	free(keys);
	free(key_lengths);
	if (res != MC_RESULT_OK) {
		php_error(E_ERROR, "Error in mc_mget(): %s", err);
		if (err != NULL)
			free(err);
		RETURN_FALSE;
	}
	RETURN_LONG(n_del);
}


/**
 * Push a record to the server, with tags. The record key is auto-generated on
 * the server and returned.
 *
 * @param 	mc		mc_connection
 * @param	value		string
 * @param	tags		array
 * @return	string
 */
PHP_FUNCTION(mc_tstack_push)
{
	struct mc_connection *mc;
	zval *zmc, *ztags, *zres, **data;
	char *rdata, *key, *err;
	struct mc_tag *tags = NULL;
	int rdata_len, res, i;
	unsigned key_len, n_tags;
	HashTable *zhash;
	HashPosition hpos;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rs|a", &zmc, &rdata, &rdata_len, &ztags) == FAILURE)
		RETURN_FALSE;
	ZEND_FETCH_RESOURCE(mc, struct mc_connection*, &zmc, -1, le_mdcached_resource_name, le_mdcached_resource);
	
	if (ztags != NULL) {
		zhash = Z_ARRVAL_P(ztags);
		n_tags = zend_hash_num_elements(zhash);
		tags = malloc(n_tags * sizeof(struct mc_tag));
		
		for (zend_hash_internal_pointer_reset_ex(zhash, &hpos), i = 0; zend_hash_get_current_data_ex(zhash, (void**) &data, &hpos) == SUCCESS; zend_hash_move_forward_ex(zhash, &hpos)) {
			long index;
			tag_key_t tkey;
			tag_value_t tvalue;

			if (zend_hash_get_current_key_ex(zhash, NULL, 0, &index, 0, &hpos) == HASH_KEY_IS_LONG) {
				if (Z_TYPE_PP(data) != IS_LONG) {
					int j;
					php_error(E_ERROR, "Array keys and values for tags must be integers");
					free(tags);
					RETURN_FALSE;
				}
				tkey = index;
				tvalue = Z_LVAL_PP(data);
				tags[i].key = tkey;
				tags[i].value = tvalue;
				i++;
			} else
				php_error(E_ERROR, "Array keys and values for tags must be integers");
		}
	}
	res = mc_tstack_push(mc, tags, n_tags, rdata, rdata_len, &key, &key_len, &err);
	free(tags);
	if (res != MC_RESULT_OK) {
		php_error(E_ERROR, "error in mc_tstack_push(): %s", err);
		if (err != NULL)
			free(err);
		RETURN_FALSE;
	}
	RETVAL_STRINGL(key, key_len, 1);
	free(key);
}


/**
 * Pops a record from the server's virtual stack identified by the tag (key & value),
 * and returns the record key-value pair as the array, with the special key '_tags'
 * containing the tags.
 *
 * @param 	mc		mc_connection
 * @param	tag_key		int
 * @param	tag_value	int
 * @return	array
 */
PHP_FUNCTION(mc_tstack_pop)
{
	struct mc_connection *mc;
	zval *zmc, *zres;
	struct mc_data_entry *mce;
	char *err;
	long tag_key, tag_value;
	int res;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rll", &zmc, &tag_key, &tag_value) == FAILURE)
		RETURN_FALSE;
	ZEND_FETCH_RESOURCE(mc, struct mc_connection*, &zmc, -1, le_mdcached_resource_name, le_mdcached_resource);
	
	if ((res = mc_tstack_pop(mc, tag_key, tag_value, &mce, &err)) != MC_RESULT_OK) {
		if (res == MC_RESULT_NOT_FOUND)
			RETURN_NULL();
		php_error(E_ERROR, "mc_tstack_pop(): %s", err);
		RETURN_FALSE;
	}
	if (mce == NULL) {
		php_error(E_ERROR, "mc_tstack_pop() returned NULL record");
		if (err != NULL)
			free(err);
		RETURN_FALSE;
	}
	zres = _php_mc_parse_single_record(mce);
	RETURN_ZVAL(zres, 0, 0);
}


/**
 * (Re)sets a record's tags
 *
 * @param	resource	mc_connection
 * @param	string		key
 * @param	array		tags
 * @return	integer
 */
PHP_FUNCTION(mc_set_tags)
{
	struct mc_connection *mc;
	zval *zmc, *ztags = NULL;
	char *key, *err = NULL;
	struct mc_tag *tags = NULL;
	unsigned n_tags;
	int key_len, ret;
	
	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rsa", &zmc, &key, &key_len, &ztags) == FAILURE)
		RETURN_FALSE;
	ZEND_FETCH_RESOURCE(mc, struct mc_connection*, &zmc, -1, le_mdcached_resource_name, le_mdcached_resource);
	if (!_php_parse_tags_from_zval(ztags, &tags, &n_tags))
		RETURN_FALSE;
	ret = mc_set_tags(mc, key, key_len, tags, n_tags, &err);
	if (tags != NULL)
		free(tags);
	if (err != NULL) {
		php_error(E_ERROR, "Error in mc_set_tags: %s", err);
		free(err);
		RETURN_FALSE;
	}
	RETURN_LONG(ret);
}

/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 * vim600: noet sw=4 ts=4 fdm=marker
 * vim<600: noet sw=4 ts=4
 */
