/*
 * MDCacheD client library for C
 * Implements interfaces defined in mc_client.h
 * (c) 2007.-2012 Ivan Voras <ivoras@gmail.com>
 * Released under the 2-clause BSDL
 * $Id: common_utils.c 448 2012-03-20 13:19:46Z ivoras $
 */

#include <sys/time.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "common_utils.h"
#include "mc_protocol.h"

/** Does a pretty hexdump of possibly human-readable data */
void
_dump_data(void *data, unsigned size) {
	unsigned char *bdata = data;
	char line[100];
	unsigned char buf[40];
	unsigned lcount = 0;
	unsigned bpos = 0;
	unsigned i,j;
	
	if (size > _DUMP_BPL) {
		sprintf(line, "%06d : ", lcount++);
		printf("** Dumping %u bytes at %p:\n", size, data);
	} else
		line[0] = 0;
	
	for (i = 0; i < size; i++) {
		bpos = i % _DUMP_BPL;
		buf[bpos] = *(bdata++);
		
		if (bpos == _DUMP_BPL-1) {
			for (j = 0; j < bpos+1; j++) {
				char bb[5];
				sprintf(bb, "%02x ", buf[j]);
				strcat(line, bb);
			}
			strcat(line, "| ");
			for (j = 0; j < bpos+1; j++) {
				unsigned char c[2];
				c[0] = buf[j];
				if (c[0] < 32 || c[0] > 127)
					c[0] = '.';
				c[1] = 0;
				strcat(line, (char*)c);
			}
			printf("%s\n", line);
			sprintf(line, "%06d : ", _DUMP_BPL * lcount++);
		}
	}
	
	if (bpos != _DUMP_BPL-1) {
/*		memset(&buf[bpos+1], 0, 20-bpos);
		bpos = 19;*/
		for (j = 0; j < bpos+1; j++) {
			char bb[5];
			sprintf(bb, "%02x ", buf[j]);
			strcat(line, bb);
		}
		if (size > _DUMP_BPL)
			for (j = 0; j < _DUMP_BPL - bpos - 1; j++)
				strcat(line, "   ");
		strcat(line, "| ");
		for (j = 0; j < bpos+1; j++) {
			unsigned char c[2];
			c[0] = buf[j];
			if (c[0] < 32 || c[0] > 127)
				c[0] = '.';
			c[1] = 0;
			strcat(line, (char*)c);
		}
		printf("%s\n", line);
	}

}


/* Returns current time as number of seconds */
double
now()
{
	struct timeval tv;
	gettimeofday(&tv, NULL);
	return (double)(tv.tv_sec)+(double)(tv.tv_usec)/1e6;
}


/**
 * Generates a pretty dump of the packet.
 */
void
_dump_data_entry(struct mc_data_entry *mde)
{
	int i;
	struct mc_tag *t;
	
	printf("key: *-------------------------------------------------------------\n");
	_dump_data(mc_data_entry_key(mde), mc_data_entry_key_size(mde));
	printf("data: -------------------------------------------------------------\ntags: ");
	_dump_data(mc_data_entry_value(mde), mc_data_entry_value_size(mde));
	for (i = 0; i < mc_data_entry_n_tags(mde); i++) {
		t = mc_data_entry_tag(mde, i);
#ifdef TAGS_64BIT
		printf("(%lld,%lld)\t", (long long)(t->key), (long long)(t->value));
#else
		printf("(%d,%d)\t", t->key, t->value);
#endif
	}
	printf("\n");
}
