/*-
 * This file is a part of the MDCacheD server
 * (c) 2007.-2011. Ivan Voras <ivoras@gmail.com>
 * Released under the 2-clause BSDL
 * $Id: logger.h 379 2011-11-10 00:09:22Z ivoras $
 */


#ifdef __cplusplus
extern "C" {
#endif

#include <syslog.h>

/* NOTE: Assume logging operations are slow and don't use them in
 * performance-critical paths. */

void setup_logger();

/**
 * Logs a message through either the internal "dummy" logger or
 * the syslog. Logger "priorities" are always those from syslog.
 * LOG_CRIT will, as a special case, terminate the process after
 * the message is logged.
 */
void mc_log(int prio, const char *msg, ...);

/**
 * Makes a simple hexpdump of the given memory region to stdout.
 * Width is the unit width for dumping (i.e. 1=byte, 4=uint32_t).
 */
void log_hexdump(const void *d, size_t len, unsigned width);

void shutdown_logger();

/**
 * Logs a critical error (which terminates the process) with a
 * human-readable message about where and what caused it.
 */
#define CRIT_ERROR(msg) mc_log(LOG_CRIT, msg " at %s (%s:%d)", __func__, __FILE__, __LINE__);


/**
 * Critical path assertion. Will abort the process similar to
 * CRIT_ERROR if the expression is not true. Is invariant to
 * NDEBUG (for NDEBUG-conscious usage user "normal" assert,
 * because CRIT_ASSERT is used to handle runtime errors).
 */
#define CRIT_ASSERT(e) if (unlikely(!(e))) CRIT_ERROR("Assertion failed: " #e);

#define TRACE(msg) mc_log(LOG_DEBUG, "##: %s at %s (%s:%d)", msg, __func__, __FILE__, __LINE__);


#ifdef __cplusplus
}
#endif

