/*-
 * Mini atomic ops header library
 * (c) 2007.-2011. Ivan Voras <ivoras@gmail.com>
 * Released under the 2-clause BSDL
 * $Id: miniatomic.h 429 2012-03-01 01:20:50Z ivoras $
 */

#ifndef _MINIATOMIC_H_
#define _MINIATOMIC_H_

/* The operations in this library are supposed to be multiprocessor-atomic. */

#ifdef __cplusplus
extern "C" {
#endif

#if __FreeBSD__ > 3
/* The atomic ops interface of this header library implements the API of the
 * FreeBSD atomic.h header. */
#include <sys/types.h>
#include <machine/atomic.h>
#elif defined(__GNUC__)
/* Mimic the API through gcc builtins - only a subset is implemented. */
#include <stdint.h>
static inline void atomic_add_32(volatile uint32_t *p, int32_t v) {
	__sync_fetch_and_add(p, v);
}
#define atomic_add_int atomic_add_32

static inline void atomic_add_64(volatile uint64_t *p, int64_t v) {
	__sync_fetch_and_add(p, v);
}

static inline uint32_t atomic_cmpset_32(volatile uint32_t *p, uint32_t v1, uint32_t v2) {
	return __sync_val_compare_and_swap(p, v1, v2);
}
#define atomic_cmpset_int atomic_cmpset_32

static inline uint64_t atomic_cmpset_64(volatile uint64_t *p, uint64_t v1, uint64_t v2) {
	return __sync_val_compare_and_swap(p, v1, v2);
}
#else
#error No atomic API can be found???
#endif

#ifdef __cplusplus
}
#endif

#endif
