/*-
 * This file is a part of the MDCacheD server
 * (c) 2007.-2012. Ivan Voras <ivoras@gmail.com>
 * Released under the 2-clause BSDL
 * $Id: misc.h 475 2012-04-08 10:27:00Z ivoras $
 */

#ifndef _MISC_H_
#define _MISC_H_

#ifndef likely
#ifdef  __GNUC__
#define likely(x)	__builtin_expect((x),1)
#define unlikely(x)	__builtin_expect((x),0)
#else
#define likely(x) (x)
#define unlikely(x) (x)
#endif
#endif

#ifndef TRUE
#define TRUE 1
#define FALSE (!TRUE)
#endif

/** Is integer a power of 2? */
#define ISPO2(x) ((x&(x-1))==0)

/** Calculate log2(v) */
static inline unsigned po2(unsigned v) {
	const unsigned int b[] = {0x2, 0xC, 0xF0, 0xFF00, 0xFFFF0000};
	const unsigned int S[] = {1, 2, 4, 8, 16};
	unsigned int r = 0;
	
	if (v & b[4]) {
		v >>= S[4];
		r |= S[4];
	}
	if (v & b[3]) {
		v >>= S[3];
		r |= S[3];
	} 
	if (v & b[2]) {
		v >>= S[2];
		r |= S[2];
	} 
	if (v & b[1]) {
		v >>= S[1];
		r |= S[1];
	} 
	if (v & b[0]) {
		v >>= S[0];
		r |= S[0];
	}
	return r;
}

#if 0
static void
_mc_debug_printf(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	vprintf(fmt, ap);
	va_end(ap);
}
#else
#define _mc_debug_printf(...)
#endif


#define _s_to_ns(s) ( (unsigned int) (s * 1000000000) )

#endif
